/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Nordic UART Service Client sample
 */

#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/printk.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

//#include <bluetooth/services/nus.h>
#include <bluetooth/services/nus_client.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>
#include "btle.h"
#include <zephyr/settings/settings.h>

#include <zephyr/drivers/uart.h>

#include <zephyr/logging/log.h>

/* SETUP THREADS */
/* size of stack area used by each thread */
#define STACKSIZE 1024

/* scheduling priority used by each thread */
#define PRIORITY 7

#define LOG_MODULE_NAME central_uart
LOG_MODULE_REGISTER(LOG_MODULE_NAME, CONFIG_LOG_MAX_LEVEL);

/* UART payload buffer element size. */
#define UART_BUF_SIZE 240

#define KEY_PASSKEY_ACCEPT DK_BTN1_MSK
#define KEY_PASSKEY_REJECT DK_BTN2_MSK

#define NUS_WRITE_TIMEOUT K_MSEC(150)
#define UART_SEND_TIMEOUT K_MSEC(100)
#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)
#define UART_RX_TIMEOUT 50000 /* Wait for RX complete event time in microseconds. */

static const struct device *uart = DEVICE_DT_GET(DT_CHOSEN(nordic_nus_uart));
static struct k_work_delayable uart_work;

K_SEM_DEFINE(nus_write_sem, 0, 1);
K_SEM_DEFINE(xltr_ble_data_received_semaphore, 0, 1);
K_SEM_DEFINE(uart_tx_complete_sem, 0, 1);


struct uart_data_t 
{
	void *fifo_reserved;
	uint8_t  data[UART_BUF_SIZE];
	uint16_t index;
};

#define DATA_RECEIVED_OVER_BLE_ARRAY_LEN 1024
static uint8_t data_received_over_ble[DATA_RECEIVED_OVER_BLE_ARRAY_LEN];
static uint16_t data_received_over_ble_in_index;
static uint16_t data_received_over_ble_out_index;

#define UART_OUT_BUF_LEN 250
static uint8_t uart_out_buf[UART_OUT_BUF_LEN];

static K_FIFO_DEFINE(fifo_uart_tx_data);
static K_FIFO_DEFINE(fifo_uart_rx_data);

static struct bt_conn *default_conn;
static struct bt_nus_client xltr_client;

static int xlrt_client_init(void);

/*
* When a  characteristic has been read by the connected XLRT this is called
* It is  setup as  a call back function. Obv.
* When data is to be give to the XLRT it is put in a characteristic
* We know that characteristic has been read when this function is called
* It is initialited in xlrt_client_init
*/
static void ble_data_sent(struct bt_nus_client *nus, uint8_t err,
					const uint8_t *const data, uint16_t len)
{
	ARG_UNUSED(nus);
	ARG_UNUSED(data);
	ARG_UNUSED(len);

	k_sem_give(&nus_write_sem);


	if (err) 
	{
		LOG_WRN("ATT error code: 0x%02X", err);
	}
}

/*
* When data comes in to the XLRT BLE UART characteristic, 
* this call back function is called
* Presumably, even though it is not used, struct of type bt_nus_client
* has to be used as this is a callback function and the underlying
* code always passes this function
*/
static uint8_t ble_data_received(struct bt_nus_client *nus,
						const uint8_t *data, uint16_t len)
{
	(void)nus;

	uint16_t i;

	for (i=0; i<len; i++)
	{
		data_received_over_ble[data_received_over_ble_in_index++] = *data++;
		if (data_received_over_ble_in_index >= DATA_RECEIVED_OVER_BLE_ARRAY_LEN)
			data_received_over_ble_in_index = 0;
	}
	k_sem_give(&xltr_ble_data_received_semaphore);

#if 0	
	int err;
	uint16_t pos;

	for (pos = 0; pos != len;) 
	{
		struct uart_data_t *tx = k_malloc(sizeof(*tx));

		if (!tx) 
		{
			LOG_WRN("Not able to allocate UART send data buffer");
			return BT_GATT_ITER_CONTINUE;
		}

		/* Keep the last byte of TX buffer for potential LF char. */
		size_t tx_data_size = sizeof(tx->data) - 1;

		if ((len - pos) > tx_data_size) 
		{
			tx->index = tx_data_size;
		} 
		else 
		{
			tx->index = (len - pos);
		}

		memcpy(tx->data, &data[pos], tx->index);

		pos += tx->index;

		/* Append the LF character when the CR character triggered
		 * transmission from the peer.
		 */
		if ((pos == len) && (data[len - 1] == '\r')) 
		{
			tx->data[tx->index] = '\n';
			tx->index++;
		}

		err = uart_tx(uart, tx->data, tx->index, SYS_FOREVER_MS);
		if (err) 
		{
			k_fifo_put(&fifo_uart_tx_data, tx);
		}
	}
#endif	

	return BT_GATT_ITER_CONTINUE;
}

static void uart_cb(const struct device *dev, struct uart_event *evt, void *user_data)
{
	ARG_UNUSED(dev);

	static size_t aborted_len;
	struct uart_data_t *buf;
	static uint8_t *aborted_buf;
	static bool disable_req;

	switch (evt->type) 
	{
	case UART_TX_DONE:
		LOG_DBG("UART_TX_DONE");
		k_sem_give(&uart_tx_complete_sem);
		break;

	case UART_RX_RDY:
		LOG_DBG("UART_RX_RDY");
		buf = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data[0]);
		buf->index += evt->data.rx.len;

		if (disable_req) 
		{
			return;
		}

		if ((evt->data.rx.buf[buf->index - 1] == '\n') ||
		    (evt->data.rx.buf[buf->index - 1] == '\r')) {
			disable_req = true;
			uart_rx_disable(uart);
		}

		break;

	case UART_RX_DISABLED:
		LOG_DBG("UART_RX_DISABLED");
		disable_req = false;

		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->index = 0;
		} else {
			LOG_WRN("Not able to allocate UART receive buffer");
			k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
			return;
		}

		uart_rx_enable(uart, buf->data, sizeof(buf->data), UART_RX_TIMEOUT);
		break;

	case UART_RX_BUF_REQUEST:
		LOG_DBG("UART_RX_BUF_REQUEST");
		buf = k_malloc(sizeof(*buf));
		if (buf) 
		{
			buf->index = 0;
			uart_rx_buf_rsp(uart, buf->data, sizeof(buf->data));
		} 
		else 
		{
			LOG_WRN("Not able to allocate UART receive buffer");
		}
		break;

	case UART_RX_BUF_RELEASED:
		LOG_DBG("UART_RX_BUF_RELEASED");
		buf = CONTAINER_OF(evt->data.rx_buf.buf, struct uart_data_t,
				   data[0]);

		if (buf->index > 0) 
		{
			k_fifo_put(&fifo_uart_rx_data, buf);
		} 
		else 
		{
			k_free(buf);
		}

		break;

	case UART_TX_ABORTED:
		LOG_DBG("UART_TX_ABORTED");
		if (!aborted_buf) {
			aborted_buf = (uint8_t *)evt->data.tx.buf;
		}

		aborted_len += evt->data.tx.len;
		buf = CONTAINER_OF(aborted_buf, struct uart_data_t,
				   data[0]);

		uart_tx(uart, &buf->data[aborted_len],
			buf->index - aborted_len, SYS_FOREVER_MS);

		break;

	default:
		break;
	}
}

static void uart_work_handler(struct k_work *item)
{
	struct uart_data_t *buf;

	buf = k_malloc(sizeof(*buf));
	if (buf) {
		buf->index = 0;
	} else {
		LOG_WRN("Not able to allocate UART receive buffer");
		k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
		return;
	}

	uart_rx_enable(uart, buf->data, sizeof(buf->data), UART_RX_TIMEOUT);
}

static int uart_init(void)
{
	int err;
	struct uart_data_t *rx;

	if (!device_is_ready(uart)) {
		LOG_ERR("UART device not ready");
		return -ENODEV;
	}

	rx = k_malloc(sizeof(*rx));
	if (rx) {
		rx->index = 0;
	} else {
		return -ENOMEM;
	}

	k_work_init_delayable(&uart_work, uart_work_handler);

	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		return err;
	}

	return uart_rx_enable(uart, rx->data, sizeof(rx->data),
			      UART_RX_TIMEOUT);
}

static void discovery_complete(struct bt_gatt_dm *dm,
			       void *context)
{
	struct bt_nus_client *nus = context;
	LOG_INF("Service discovery completed");

	bt_gatt_dm_data_print(dm);

	bt_nus_handles_assign(dm, nus);
	bt_nus_subscribe_receive(nus);

	bt_gatt_dm_data_release(dm);
}

static void discovery_service_not_found(struct bt_conn *conn,
					void *context)
{
	LOG_INF("Service not found");
}

static void discovery_error(struct bt_conn *conn,
			    int err,
			    void *context)
{
	LOG_WRN("Error while discovering GATT database: (%d)", err);
}

struct bt_gatt_dm_cb discovery_cb = {
	.completed         = discovery_complete,
	.service_not_found = discovery_service_not_found,
	.error_found       = discovery_error,
};

static void gatt_discover(struct bt_conn *conn)
{
	int err;

	if (conn != default_conn) {
		return;
	}

	err = bt_gatt_dm_start(conn,
			       BT_UUID_XLRT_SERVICE,
			       &discovery_cb,
			       &xltr_client);
	if (err) {
		LOG_ERR("could not start the discovery procedure, error "
			"code: %d", err);
	}
}

static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params)
{
	if (!err) {
		LOG_INF("MTU exchange done");
	} else {
		LOG_WRN("MTU exchange failed (err %" PRIu8 ")", err);
	}
}

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		LOG_INF("Failed to connect to %s, 0x%02x %s", addr, conn_err,
			bt_hci_err_to_str(conn_err));

		if (default_conn == conn) {
			bt_conn_unref(default_conn);
			default_conn = NULL;

			err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
			if (err) {
				LOG_ERR("Scanning failed to start (err %d)",
					err);
			}
		}

		return;
	}

	LOG_INF("Connected: %s", addr);

	static struct bt_gatt_exchange_params exchange_params;

	exchange_params.func = exchange_func;
	err = bt_gatt_exchange_mtu(conn, &exchange_params);
	if (err) {
		LOG_WRN("MTU exchange failed (err %d)", err);
	}

	err = bt_conn_set_security(conn, BT_SECURITY_L2);
	if (err) {
		LOG_WRN("Failed to set security: %d", err);

		gatt_discover(conn);
	}

	err = bt_scan_stop();
	if ((!err) && (err != -EALREADY)) {
		LOG_ERR("Stop LE scan failed (err %d)", err);
	}
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Disconnected: %s, reason 0x%02x %s", addr, reason, bt_hci_err_to_str(reason));

	if (default_conn != conn) {
		return;
	}

	bt_conn_unref(default_conn);
	default_conn = NULL;

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)",
			err);
	}
}

static void security_changed(struct bt_conn *conn, bt_security_t level,
			     enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (!err) {
		LOG_INF("Security changed: %s level %u", addr, level);
	} else {
		LOG_WRN("Security failed: %s level %u err %d %s", addr, level, err,
			bt_security_err_to_str(err));
	}

	gatt_discover(conn);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
	.security_changed = security_changed
};

static void scan_filter_match(struct bt_scan_device_info *device_info,
			      struct bt_scan_filter_match *filter_match,
			      bool connectable)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	LOG_INF("Filters matched. Address: %s connectable: %d",
		addr, connectable);
}

static void scan_connecting_error(struct bt_scan_device_info *device_info)
{
	LOG_WRN("Connecting failed");
}

static void scan_connecting(struct bt_scan_device_info *device_info,
			    struct bt_conn *conn)
{
	default_conn = bt_conn_ref(conn);
}

/*
 * Sets up two callback functions for the BLE UART
 * ble_data_received is for when data comes in from the XLRT
 * ble_data_sent is  set up for when BLE data has been sent out 
 * and read by the XLRT. It is set up as a clalback function via a 
 * tortuous call to bt_nus_client_init which  is in nus_client.c
 * and which should probab;y be copied to a new file with 
 * lots of renaming of functions 


*/
static int xlrt_client_init(void)
{
	int err;
	struct bt_nus_client_init_param init = 
	{
		.cb = 
		{
			.received = ble_data_received,
			.sent = ble_data_sent,
		}
	};

	err = bt_nus_client_init(&xltr_client, &init);
	if (err) 
	{
		LOG_ERR("NUS Client initialization failed (err %d)", err);
		return err;
	}

	LOG_INF("NUS Client module initialized");
	return err;
}

BT_SCAN_CB_INIT(scan_cb, scan_filter_match, NULL, scan_connecting_error, scan_connecting);

static int scan_init(void)
{
	int err;
	struct bt_scan_init_param scan_init = 
	{
		.connect_if_match = 1,
	};

	bt_scan_init(&scan_init);
	bt_scan_cb_register(&scan_cb);

	err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_XLRT_SERVICE);
	if (err) 
	{
		LOG_ERR("Scanning filters cannot be set (err %d)", err);
		return err;
	}

	err = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, false);
	if (err) 
	{
		LOG_ERR("Filters cannot be turned on (err %d)", err);
		return err;
	}

	LOG_INF("Scan module initialized");
	return err;
}


static void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing cancelled: %s", addr);
}


static void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing completed: %s, bonded: %d", addr, bonded);
}


static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_WRN("Pairing failed conn: %s, reason %d %s", addr, reason,
		bt_security_err_to_str(reason));
}

static struct bt_conn_auth_cb conn_auth_callbacks = {
	.cancel = auth_cancel,
};

static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
	.pairing_complete = pairing_complete,
	.pairing_failed = pairing_failed
};

//int main(void)
int uart_in_thread(void)
{
	int err;
	int bytes_remaining_from_uart_data_in;
	struct uart_data_t xlrt_ble_data_out;
	int number_of_bytes_to_put_in_ble_buffer;
	int space_left_in_btle_buffer;
	

	data_received_over_ble_in_index = 0;
	data_received_over_ble_out_index = 0;

	err = bt_conn_auth_cb_register(&conn_auth_callbacks);
	if (err) {
		LOG_ERR("Failed to register authorization callbacks.");
		return 0;
	}

	err = bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
	if (err) 
	{
		printk("Failed to register authorization info callbacks.\n");
		return 0;
	}

	err = bt_enable(NULL);
	if (err) {
		LOG_ERR("Bluetooth init failed (err %d)", err);
		return 0;
	}
	LOG_INF("Bluetooth initialized");

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	err = uart_init();
	if (err != 0) {
		LOG_ERR("uart_init failed (err %d)", err);
		return 0;
	}

	err = scan_init();
	if (err != 0) {
		LOG_ERR("scan_init failed (err %d)", err);
		return 0;
	}

	err = xlrt_client_init();
	if (err != 0) {
		LOG_ERR("nus_client_init failed (err %d)", err);
		return 0;
	}

	printk("Starting Bluetooth Central UART example\n");

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)", err);
		return 0;
	}

	LOG_INF("Scanning successfully started");

	
	xlrt_ble_data_out.index = 0;
	

	for (;;) 
	{
		/* Wait indefinitely for data to be sent over Bluetooth */
		struct uart_data_t *buf = k_fifo_get(&fifo_uart_rx_data, K_FOREVER); //K_FOREVER);
		
			// How much space is left in the UART buffer? 
			space_left_in_btle_buffer = sizeof(xlrt_ble_data_out.data) - xlrt_ble_data_out.index;
			// What's smaller, space left in btle buffer or amount or incoming data from the uart
			// If there are fewer bytes available in the BLE output buffer than there are in
			// the received data buffer all is copied.
			// Otherwise, only the amount of space left in the outgoing BLE buffer is copied
			number_of_bytes_to_put_in_ble_buffer = MIN(space_left_in_btle_buffer, buf->index);
			int loc = 0;

			bytes_remaining_from_uart_data_in = buf->index;
			while (number_of_bytes_to_put_in_ble_buffer > 0) 
			{
				// Copy from uart rx data buffer into ble out buffer. Number of bytes been calculated already
				memcpy(&xlrt_ble_data_out.data[xlrt_ble_data_out.index], &buf->data[loc], number_of_bytes_to_put_in_ble_buffer);
				xlrt_ble_data_out.index += number_of_bytes_to_put_in_ble_buffer;
				loc += number_of_bytes_to_put_in_ble_buffer;

				number_of_bytes_to_put_in_ble_buffer = MIN(sizeof(xlrt_ble_data_out.data), buf->index - loc);

				/*if (xlrt_ble_data_out.index >= sizeof(xlrt_ble_data_out.data) ||
				(xlrt_ble_data_out.data[xlrt_ble_data_out.index - 1] == '\n') ||
				(xlrt_ble_data_out.data[xlrt_ble_data_out.index - 1] == '\r'))  */
				if ( (number_of_bytes_to_put_in_ble_buffer == 0) || (xlrt_ble_data_out.index >= sizeof(xlrt_ble_data_out.data)))
				{
						err = bt_nus_client_send(&xltr_client, xlrt_ble_data_out.data, xlrt_ble_data_out.index);
						if (err) 
						{
							LOG_WRN("Failed to send data over BLE connection"
								"(err %d)", err);
						}

						err = k_sem_take(&nus_write_sem, NUS_WRITE_TIMEOUT);
						if (err) 
						{
							LOG_WRN("NUS send timeout");
						}

					xlrt_ble_data_out.index = 0;
				}

				//number_of_bytes_to_put_in_ble_buffer = MIN(sizeof(xlrt_ble_data_out.data), buf->index - loc);
			}

			k_free(buf);
		
	}
		
}
void uart_out_thread(void)
{
	uint16_t uart_out_index;
	uart_out_index = 0;
	// Need to make sure this sempahore is available for the first lot of data in
	k_sem_give(&uart_tx_complete_sem); 
	while (1)
	{
		k_sem_take(&xltr_ble_data_received_semaphore, K_FOREVER);
		k_sem_take(&uart_tx_complete_sem, UART_SEND_TIMEOUT);
		uart_out_index = 0;
		if (data_received_over_ble_in_index != data_received_over_ble_out_index) // && (uart_out_index < UART_OUT_BUF_LEN))
		{				
			while ((data_received_over_ble_in_index != data_received_over_ble_out_index) && (uart_out_index < UART_OUT_BUF_LEN))
			{
				uart_out_buf[uart_out_index++] = data_received_over_ble[data_received_over_ble_out_index++];
				if (data_received_over_ble_out_index >= DATA_RECEIVED_OVER_BLE_ARRAY_LEN)
					data_received_over_ble_out_index = 0;					
			}
			uart_tx(uart, uart_out_buf, uart_out_index, SYS_FOREVER_MS);
		}
		
	}

}

K_THREAD_DEFINE(uart_in_thread_id, STACKSIZE, uart_in_thread, NULL, NULL, NULL,
		PRIORITY, 0, 0);
K_THREAD_DEFINE(uart0_out_thread_id, STACKSIZE, uart_out_thread, NULL, NULL, NULL,
		PRIORITY, 0, 0);